/************************************************
* Utilisation de GMPLIB sous licence LGPL
-------------------------------------------------
* Utilisation de la librairie VIFF sous licence
* LGPL pour les fonction share et recombine.
* Conversion de Python  C++ faite par
* Ruben BOKOBZA (MORPHO) le 04 septembre 2013
************************************************/

#include <iostream>
#include <sstream>
#include <string>
#include <assert.h>
#include <vector>
#include <gmp.h>
#include <gmpxx.h>
#include <time.h>
#include "FieldElement.hpp"
#include "ShamirSharing.hpp"

mpz_class ShamirSharing::prime("175881368863475933545584710859272828821808893362503327722984450798255471312301");
unsigned int ShamirSharing::numBits = 256;

using namespace std;

/**
* Codage d'un tableau d'octets dans un entier
* (Octets -> Bits -> Entier)
* Le rsultat est une liste d'entiers de numBits
* Exemple : 
* Codage de 1001000101001010101010010010001110000 :
* Dcoupage (exemple a numBits = 16) : On coupe en partant des bits de poids faible :
* 10010 0010100101010101 0010010001110000
* puis convertissage en entiers : 18, 10581, 9328
*/
vector<mpz_class> ShamirSharing::bytesToInt(char* x, int n, unsigned int numBits){
	stringstream s;
	vector<mpz_class> res(n*8/numBits + 1);
	mpz_t tmp;
	mpz_init(tmp);
	int i, cp = 0;
	
	for (unsigned int i = 0 ; i < (numBits-((n*8)%numBits)) ; i++){
		s << "0";
	}
	// Pour chaque octet
	for (i = 0 ; i < n ; i++){
		// On extrait les bits 
		int o = int(x[i])+127;
		for (int j = 7 ; j >= 0; j--){
			s << ((o >> j) & 1);
		}
		// Si on arrive a la fin d'un bloc, on stocke le buffer
		if ((n-i-1) % (numBits/8) == 0 ){
			res[cp++] = mpz_class(s.str(),2);
			s.str("");
		}
	}
	return res;
} 


/**
* Dcodage int -> Octets
* Exemple : 
* Dcodage de 18, 10581, 9328
* 18 Est le premier donc on sait qu'il commence par 1 : 
* On va le convertir en octets : 10010 -> 00010010
* - On rajoute des 0 devant pour completer l'octet incomplet (on ne complete pas jusqu' numBits!)
* 10581 et 9328 suivent, on sait qu'ils sont complets : on les complete jusqu'a numBits
* 10100101010101 -> 0010100101010101, 10010001110000 -> 0010010001110000
*/
char* ShamirSharing::intToBytes(mpz_class x, int* size,unsigned int numBits, bool first){
	char* r = (char*) malloc(numBits/8*sizeof(*r));
	for (unsigned int i = 0 ; i < numBits/8 ; i++ ) r[i] = 0 ;
	int sum;
	string bin = mpz_get_str(NULL,2,x.get_mpz_t());

	// Padding - cf plus haut
	if (!first){
		int sizeTmp = bin.size();
		for (unsigned int i = 0 ; i < (numBits-sizeTmp) ; i++){
			bin = "0" + bin;
		}
	}else{
		int sizeTmp = bin.size();
		for (int i = 0 ; i < (8-(sizeTmp%8))%8 ; i++){
			bin = "0" + bin;
		}
	}

	// Il ne reste qu'a recopier les bits dans des octets
	int cp = 0;
	for (unsigned int i = 0 ; i < bin.size() ; i += 8){
		sum = 0;
		for(int j = 0; j < 8; j++){
			if (i + j < bin.size()){
				sum = sum << 1 | (bin.substr(i,8).at(j) - '0');
			}
		}
		r[cp++] = (sum-127==128)?char(-(sum-127)):char(sum-127);
	}

	r = (char*) realloc(r,cp*sizeof(*r));
	*size = cp;
	return r;
}


/**
* Entier -> Shares (liste de paires (id,share))
* VIFF : Licence LGPL
*/
vector< pair<int,FieldElement> > ShamirSharing::share(FieldElement secret, int threshold, int num_players){
	assert (threshold > 0 && threshold < num_players);
	gmp_randclass random(gmp_randinit_default);

	FieldElement coeff[threshold+1];
	mpz_class prime = secret.getModulo();

	// Construction du polynome : (threshold) coefs alatoires
	// P = c[0](=secret) + c[1]*X + ... + c[t]*X^t
	coeff[0] = secret;
	for (int i = 1 ; i <= threshold ; i++){
		coeff[i] = FieldElement(random.get_z_range(prime),prime);
	}

	// Evaluation en "id"
	// P(X) = c[0] + X*(c[1] + X*(... (+ X*c[t])...)
	// P(id) = c[0] + id*(c[1] + id*(... (+ id*c[t])...)
	// Stockage sous la forme (id,P(id))
	vector< pair<int,FieldElement> > shares(num_players);
	for (int i = 1 ; i <= num_players ; i++){
		shares[i-1] = make_pair(i,coeff[threshold]);
		for (int j = threshold - 1; j >=0 ; j--){
			shares[i-1].second = coeff[j] + FieldElement(mpz_class(i),prime) * shares[i-1].second;
		}
	}

	return shares;
}

/*
* Shares (liste de paires (id,share)) -> Entier
* VIFF : Licence LGPL
*/
FieldElement ShamirSharing::recombine(vector< pair<int,FieldElement> > shares){
	mpz_class prime = shares[0].second.getModulo();
	FieldElement sum(prime);
	FieldElement prod(prime);
	
	// Interpolation de Lagrange (en 0)
	// Etant donn une liste (i,P(i)) : Trouver P(0)
	// sum_{i=0}^{t+1}{P(i)*prod}
	for (unsigned int i = 0 ; i < shares.size() ; i++){
		prod = shares[i].second;
		// prod_{j=0,j!=i}^{t+1}{(-x_j)/(x_i-x_j)}
		for (unsigned int j = 0 ; j < shares.size() ; j++){
			if (i != j){
				prod *= FieldElement(-(shares[j].first),prime) / FieldElement(shares[i].first-shares[j].first,prime);
			}
		}
		sum += prod;
	}

	return sum;
}